/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001, 2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "game.hxx"
#include "room.hxx"
#include "hexpage.hxx"


Game::Game(SpectrumMemory *mem)
{
	m_mem = mem;
	m_filename[0] = 0;
	m_hexPage = new HexEditPage(m_mem);
	m_rc = NULL;
	m_spriteMap = NULL;
	m_room = 0;
}

Game::~Game()
{
	delete m_hexPage;
	delete m_mem;
	clearRefCounts();
	clearSpriteMap();
}

char *Game::getID()
{
	char *name, *pname;
	
	name  = m_filename;
	pname = strrchr(name, '/');
	if (pname) name = pname + 1;
#ifdef _WINDOWS
	pname = strrchr(name, '\\');
	if (pname) name = pname + 1;
	pname = strrchr(name, ':');
	if (pname) name = pname + 1;
#endif
	sprintf(m_namebuf, "%-*.*s", VIDEO_W / CHAR_W, VIDEO_W / CHAR_W, name);
	return m_namebuf;
}





int Game::editMain()
{
	m_keyFilter = setAppKeyFilter(this);
	JswNotebook p(0, 0, VIDEO_W, VIDEO_H);
	p.setTitle("Screens");
	addPages(&p);
	int rv = p.doModal();
	deletePages();
	setAppKeyFilter(m_keyFilter);
	return rv;
}


int Game::onKeyDown(int keysym)
{
	switch(keysym)
	{
		case ZXK_F5:	return onTest();
		case ZXK_F2:	return onSave();
	}

        if (m_keyFilter) return m_keyFilter->onKeyDown(keysym);
        return -1;
}


int Game::onKeyUp(int keysym)
{
	if (m_keyFilter) return m_keyFilter->onKeyUp(keysym);
	return -1;
}


int Game::onTest(void)
{
	// Compress data (if the game uses compressed data)
	int rv = compress();
	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_CANCEL) return ZXE_CONTINUE;

	char temp_name[PATH_MAX];
	FILE *fp = openTemp(temp_name, ".z80", "w+b");
	if (!fp) 
	{
		alert("Test game", "Cancel", 
			"Couldn't save temporary file:\n%s", temp_name);
		return ZXE_CONTINUE;
	}
	m_mem->saveZ80(temp_name, fp);
	fclose(fp);

	return executeZ80(temp_name);
}

int Game::onSave(void)
{
        int rv;

	// Compress data (if the game uses compressed data)
	rv = compress();
	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_CANCEL) return ZXE_CONTINUE;

	if (!m_filename[0]) 
	{
		return alert("Save game", "Cancel", 
				"No filename provided. Cannot save.");
	}
        FILE *fp = fopen(m_filename, "wb");
        if (!fp)
        {
		return alert("Save game", "Cancel", 
				"Failed to open file:\n%s", m_filename);
        }
        SpectrumMemory *mem = getMem();

        rv = mem->toDisc(m_filename, fp);
        fclose(fp);
        if (rv >= ZXE_QUIT) return rv;

	rv = alert("Game saved", "OK"); 
	if (rv >= ZXE_QUIT) return rv;
	return ZXE_CONTINUE;
}

int Game::afterLoadFix(void)
{
	return ZXE_CONTINUE;
}

void Game::addPages(VideoNotebook *n)
{
	n->addPage(m_hexPage);	
}


void Game::deletePages()
{

}


int Game::decompress()
{
	return ZXE_CONTINUE;
}

int Game::compress()
{
	return ZXE_CONTINUE;
}


void Game::clearRefCounts()
{
	while(m_rc)
	{
		RefCount *rc2 = m_rc->next;
		delete m_rc;
		m_rc = rc2;
	}
}

void Game::addRef(int id)
{
	RefCount *rc = m_rc;
	while(rc)
	{
		if (rc->id == id) 
		{
			++rc->count;
			return;
		}
		rc = rc->next;
	}
	rc = new RefCount;
	rc->next = m_rc;
	rc->id = id;
	rc->count = 1;
	m_rc = rc;
}

void Game::unRef(int id)
{
	RefCount *rc = m_rc;
	while(rc)
	{
		if (rc->id == id) 
		{
			if (rc->count > 0)
				--rc->count;
			return;
		}
		rc = rc->next;
	}
}



int Game::getRefCount(int id)
{
	RefCount *rc = m_rc;
	while(rc)
	{
		if (rc->id == id) return rc->count;
		rc = rc->next;
	}
	return 0;
}

void Game::dumpRefs()
{
	RefCount *rc = m_rc;
	while(rc)
	{
		printf("id: %c %06x  count: %d\n",
			(rc->id >> 24), (rc->id & 0xFFFFFF), rc->count);
		rc = rc->next;
	}
}


void Game::exportReferencedSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, int w, int h)
{
	int n, x, y;
	xmlNodePtr node;
	RefCount *rc = m_rc;
	int spriteLen = (w * h) / 8;
	char *buf = new char[spriteLen * 3 + 10];
	const jswByte *sprite;

	while(rc)
	{
		if ((rc->id & 0xFF000000) == SPRITE(0))
		{
			int rm = (rc->id >> 16) & 0xFF;
			sprite = getSprite(rc->id);
			if (sprite != NULL)
			{
				node = xmlNewNode(ns, (xmlChar *)"sprite");
				sprintf(buf, "%06x", rc->id & 0xFFFFFF);
				xmlSetProp(node, (xmlChar *)"base", (xmlChar *)buf);
				sprintf(buf, "%d", w);
				xmlSetProp(node, (xmlChar *)"w", (xmlChar *)buf);
				sprintf(buf, "%d", h);
				xmlSetProp(node, (xmlChar *)"h", (xmlChar *)buf);
				buf[0] = 0;
				n = 0;
				for (y = 0; y < h; y++)
				{
					for (x = 0; x < (w/8); x++)
					{
						char tmp[4];
						sprintf(tmp, "%02x", sprite[n++]);
						strcat(buf, tmp);
					}
					strcat(buf, "\n");
				}
				xmlAddChild(node, xmlNewText((xmlChar *)buf));
				xmlAddChild(parent, node);
			}
		}
		rc = rc->next;
	}
	delete buf;
}

void Game::clearSpriteMap()
{
	if (m_spriteMap) delete m_spriteMap;
	m_spriteMap = NULL;
}

void Game::addSpriteRef(int fileid, int memid)
{
	int n;

//	printf("addSpriteRef(%x,%x)\n", fileid, memid);
	if (!m_spriteMap)
	{
		m_spriteMap = new unsigned[10];
		memset(m_spriteMap, 0, 10 * sizeof(unsigned));
		m_spriteMapLen = 10;
	}
	for (n = 0; n < m_spriteMapLen; n += 2)
	{
		if (m_spriteMap[n] == fileid)
		{
			m_spriteMap[n + 1] = memid;
			return;
		}
		if (m_spriteMap[n + 1] == 0)
		{
			m_spriteMap[n    ] = fileid;
			m_spriteMap[n + 1] = memid;
			return;
		}
	}
	unsigned *os = m_spriteMap;

	m_spriteMap = new unsigned[2 * m_spriteMapLen];
	memset(m_spriteMap, 0,  2 * m_spriteMapLen * sizeof(unsigned)); 
	memcpy(m_spriteMap, os,     m_spriteMapLen * sizeof(unsigned));
	delete [] os;
	m_spriteMap[m_spriteMapLen]     = fileid;
	m_spriteMap[m_spriteMapLen + 1] = memid;
	m_spriteMapLen *= 2;	
}


int Game::lookupSpriteRef(int fileid)
{
	int n;

	if (!m_spriteMap) return 0;
	
	for (n = 0; n < m_spriteMapLen; n += 2)
	{
		if (m_spriteMap[n] == fileid)
		{
//			printf("lookupSpriteRef(%x)=%x\n", fileid, m_spriteMap[n+1]);
			return m_spriteMap[n + 1];
		}
	}
	return 0;
}



bool Game::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
        xmlNodePtr cur;
	clearSpriteMap();
	int memid, fileid, w, h, x, hexv;
	jswByte *buf;
	char *t, *t1, hex[3];

	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
		if (!s) continue;
		fileid = 0;
		sscanf(s, "%x", &fileid);
		xmlFree(s);
		s = (char *)xmlGetProp(cur, (xmlChar *)"w");
		if (!s) continue;
		w = atoi(s);
		xmlFree(s);
		if (!w) continue;	
		s = (char *)xmlGetProp(cur, (xmlChar *)"h");
		if (!s) continue;
		h = atoi(s);
		xmlFree(s);
		if (!h) continue;	
	
		t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if (!t) continue;

		buf = new jswByte[w*h/8];
		t1 = t;
		x = 0;
		hex[0] = 0;
		while (t1[0])
		{
			if (!isxdigit(t1[0]))
			{
				++t1;
				continue;
			}
			if (hex[0] == 0) 
			{
				hex[0] = *t1;
				hex[1] = 0;
				++t1;
				continue;
			}
			hex[1] = *t1;
			hex[2] = 0;
			++t1;
			hexv = 0;
			sscanf(hex, "%x", &hexv);
			hex[0] = 0;
			buf[x++] = hexv;
			if (x >= (w*h/8)) break;
		}
		xmlFree(t);
		/* buf is now the parsed sprite */
		memid = addSprite(fileid, w, h, buf);
		if (memid >= 0)
		{
			addSpriteRef(fileid, memid);
		}
		delete buf;
	}
	return false;	
}


/* Generic Game doesn't know what to do with imported sprites */
int Game::addSprite(int fileid, int w, int h, unsigned char *data)
{
	return -1;
}


jswByte *Game::getFont(int *len, int *firstc)
{
	int addr = (m_mem->peekW(23606));

	if (addr > 16384)
	{
		if (firstc) *firstc = 32;
		if (len) *len = 96;
		return m_mem->memoryAt(addr+256);
	}
	return NULL;
}


Room *Game::getRoomClass(int room)
{
	return NULL;
}


JswRoomList *Game::getRoomList(void)
{
	return NULL;
}


void Game::addFixedRefs()
{

}


bool Game::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent, int *roomMap)
{
	return true;
}


jswByte *Game::getSprite(int id)
{
	return getMem()->memoryAt(id & 0xFFFF);
}
