/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2000-2001,2004-6  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "jswgame.hxx"
#include "j128guard.hxx"
#include "j128glist.hxx"
#include "guardedit.hxx"
#include "j128gedit.hxx"
#include "spritelist.hxx"
#include "room.hxx"
#include <assert.h>

// Sprite import is amazingly fiddly. Turn on this define to produce
// debugging messages during it
#define DEBUG_SPRITE_IMPORT 0

JswGame::JswGame(SpectrumMemory *mem) : Game(mem)
{
	m_upgrade = NULL;
	m_itemXY    = 0xA500;
	m_itemCount = 0xA3FF;
	m_itemBase  = 0xA400;
	m_itemStat  = 0xA400;
	m_itemMask  = 0x3F;
	m_guardianMask= 0x7F;
	m_guardsBase  = 0xA000;
	m_guardsEnd   = 0xA600;
	m_spriteBase  = 0x9C00;

	/* Initial PC for JSW48 */
	m_mem->poke(0x5BFE, 0xCA);
	m_mem->poke(0x5BFF, 0x87);
	m_mem->clearDirty();
	m_roomCache = NULL;
	m_roomCacheSize = 0;
	m_roomNameOffset = 128;
	m_isp = NULL;
	m_sprites = NULL;
}

JswGame::~JswGame()
{
	if (m_roomCache)
	{
		for (int n = 0; n < m_roomCacheSize; n++) 
			if (m_roomCache[n]) delete m_roomCache[n];
 		delete [] m_roomCache;
	}
	if (m_isp)     delete [] m_isp;
	if (m_sprites) delete [] m_sprites;
}


/* Check for Ian Collier's "avoid infinite death" patch. If it's present,
 * start position is set by room. As you can see from the function name,
 * I wasn't sure whether to treat it as a full "mode" or not.
 *
 * Need to back this patch out if going to JSW128.
 * XXX Disallow applying it in Geoff or Andy Mode. 
 */
bool JswGame::imcMode(void)
{
	return (m_mem->peek(0x8C01) == 0x3A); 
}

/* Check for Andrew Broad's accents patch; this changes how characters 
 * 0-31 are drawn */
int JswGame::canAccent(void)
{
	unsigned addr = m_mem->peekW(0x9684);
      
	if (addr == 0x9691) return 0;	/* Standard printing function */	

	return m_mem->peek(addr + 10);	/* Sprite page for extra chars */
}


int JswGame::getStartRoom(void)
{
	if (imcMode()) return m_mem->peek(0x8420);
	return m_mem->peek(0x87EB);
}

void JswGame::setStartRoom(int room)
{
	if (imcMode()) m_mem->poke(0x8420, room);
	else m_mem->poke(0x87EB, room);
}

int JswGame::getMariaRoom(void)
{
        return m_mem->peek(0x9538);
}

void JswGame::setMariaRoom(int room)
{
        m_mem->poke(0x9538, room);
	checkMariaToilet();
}

int JswGame::getToiletRoom(void)
{
        return m_mem->peek(0x9588);
}

void JswGame::setToiletRoom(int room)
{
        m_mem->poke(0x959E, room);
	m_mem->poke(0x9588, room);
	checkMariaToilet();
}

int JswGame::chooseGuardianType(int *t)
{
	int rv;

	if (supportsExtGuards())
	{
		static int types[] = { 0, 1, 2, 3, 4, 5, 6, 7, 9, 13, 14 };
 		VideoMenu vm128("Guardian type",
				"Blank",                        // 0
				"Horizontal",                   // 1
				"Vertical",                     // 2
				"Rope",                         // 3
				"Arrow",                        // 4
				"Diagonal NW/SE",               // 5
				"Diagonal NE/SW",               // 6 
				"Vertical multicolour",         // 7
				"Horizontal multicolour",       // 8 -> 9
				"NW/SE multicolour",            // 9 -> 13
				"NE/SW multicolour",            // 10 -> 14
				NULL);
		rv = vm128.doModal();
		if (t) 
		{
			*t = types[vm128.getSelected()];	
		}
	}
	else
	{
        	VideoMenu vm48("Guardian type", "Blank", "Horizontal", 
				"Vertical", "Rope", "Arrow", NULL);
		rv = vm48.doModal();
		if (t) *t = vm48.getSelected();
	}
	return rv;
}





void JswGame::itemDelete(int index)
{
	int itemmin = m_mem->peek(m_itemCount);

	if (itemmin == 0xFF) return;	// Can't delete last item
	// Move all items up one
	for (int n = index; n > itemmin; n--)
	{
		m_mem->poke(m_itemBase + n, m_mem->peek(m_itemBase + n - 1));
		m_mem->poke(m_itemXY   + n, m_mem->peek(m_itemXY   + n - 1));
		m_mem->poke(m_itemStat + n, m_mem->peek(m_itemStat + n - 1));
	}
	m_mem->poke(m_itemCount, itemmin + 1);
}


void JswGame::itemInsert(int room, int x, int y)
{
	int itemmin = m_mem->peek(m_itemCount);
	
	if (!itemmin) return;	// All items taken

	--itemmin;
	y *= 2;
	m_mem->poke(m_itemStat + itemmin, ((y & 0x80) | 0x40)); 
	m_mem->poke(m_itemXY   + itemmin, (x & 31) | ((y & 0x70) << 1));
	m_mem->poke(m_itemBase + itemmin,
		  (m_mem->peek(m_itemBase + itemmin) & ~m_itemMask) | room); 
	m_mem->poke(m_itemCount, itemmin);
}


int  JswGame::itemPos(int index, int *room, int *x, int *y)
{
	int oy = m_mem->peek(m_itemStat + index) & 0x80;
	int ox = m_mem->peek(m_itemXY   + index);

	if (index < m_mem->peek(m_itemCount)) return -1;

	*room = m_mem->peek(m_itemBase + index) & m_itemMask;


	*x = (ox & 0x1F);
	*y = ((ox & 0xE0) >> 2)	| (oy >> 1); 
//	printf("%d: %02x %02x -> %d %d %d\n",
//		index, oy, ox, *room, *x, *y);
	return 0;
}


int  JswGame::itemAt(int room, int x, int y)
{
        int itemmin = m_mem->peek(m_itemCount);
        int n, room1, x1, y1;

	for (n = itemmin; n <= 0xFF; n++)
	{
		if (!itemPos(n, &room1, &x1, &y1))
		{
			if (x1 == x && y1 == y && room1 == room) 
				return n;
		}

	}
	return -1;
}

int JswGame::guardTableCount(void)
{
	return 1;
}


int  JswGame::chooseGuardTable(int *ng)
{
	*ng = m_guardsBase;
	return ZXE_OK;
}

jswByte *JswGame::getGuards(int room)
{
        return m_mem->memoryAt(getGuardBase(room));
}

int JswGame::getGuardBase(int room)
{
	return m_guardsBase;
}

int JswGame::chooseGuard(VideoBitmap *bm, jswByte *table, int *ng, int bg)
{
	Jsw128GuardList gl(bm, table, m_mem, bg, getGuardianMask());

	int rv = gl.doModal();

	if (rv >= ZXE_CANCEL) return rv;
	*ng = gl.getSelected();	
	return rv;
}

int JswGame::allowSuperJump(void)
{
	return 0;
}


int JswGame::allowSpriteOverride(void)
{
	if (m_mem->peek(0x9654) == 0x20 &&
            m_mem->peek(0x9655) == 0x84) return 0;	
	return m_mem->peek(0x9654);
}


int JswGame::decodeWillySprite(int room, int sprite)
{
	int j = allowSpriteOverride();
	
	if (j) return sprite;
	
	if (room == m_mem->peek(0x9657)) return m_mem->peek(0x965B);
	else				 return m_mem->peek(0x9652);
}

int JswGame::encodeWillySprite(int room, int sprite)
{
	int j = allowSpriteOverride();

	if (j) return sprite;

	return 0;
}



void JswGame::getStartPos(int *x, int *y)
{
	if (imcMode()) 
	{
		*x = m_mem->peek(0x80EE) & 0x1F;
		*y = m_mem->peek(0x80ED);
	}
	else
	{
		*x = m_mem->peek(0x87F0) & 0x1F;
		*y = m_mem->peek(0x87e6);
	}
}


void JswGame::setStartPos(int x, int y)
{
	if (imcMode()) 
	{
		m_mem->poke(0x80EE, ((y << 1) & 0xE0) | (x & 0x1F));
		m_mem->poke(0x80EF, (0x5C | (y >> 7)));
		m_mem->poke(0x80ED, y);
	}
	else
	{
		m_mem->poke(0x87F0, ((y << 1) & 0xE0) | (x & 0x1F));
		m_mem->poke(0x87F1, (m_mem->peek(0x87F1) & 0xFE) | (y >> 7));
		m_mem->poke(0x87E6, y);
	}
}

void JswGame::getMariaPos(int *x, int *y)
{
	int hl = m_mem->peek(0x955F) + 256 * m_mem->peek(0x9560);
	int y1, y2, y3;

	*x =  hl & 0x1F;
	
	hl &= 0x0FE0;
        y1 = (hl >>  8) & 7;
        y2 = (hl >>  5) & 7;
        y3 = (hl >> 11) & 3;

        *y = 2*(y1 + 8*y2 + 64*y3);
}

void JswGame::setMariaPos(int x, int y)
{
	int hl = m_mem->peek(0x955F) + 256 * m_mem->peek(0x9560);
	int y1, y2, y3;
	y &= 0xF0;
	int w = (2*y) + x;
	
	hl = (hl & ~(0xFFF)) | x;
	y1 = ((y >> 1) & 7) << 8;
	y2 = ((y >> 4) & 7) << 5; 
	y3 = ((y >> 7) & 3) << 11;
	hl |= (y1+y2+y3);

	m_mem->pokeW(0x955F, hl);

	hl = 0x5C00 + w;
        m_mem->pokeW(0x956d, hl);
	hl += 32;
        m_mem->poke(0x9573, hl);
}


void JswGame::getToiletPos(int *x, int *y)
{	
	*x = m_mem->peek(0x95B9) & 0x1F;
	*y = m_mem->peek(0x95B6);
}

void JswGame::setToiletPos(int x, int y)
{
        int hl;

	m_mem->poke(0x95B9, x);
	m_mem->poke(0x95B6, y);

	y *= 2;
	y &= 0x1E0;
	m_mem->poke(0x958E, x | (y & 0xE0));

        hl = 0x5C00 + y + x;
        m_mem->pokeW(0x95c2, hl);
        hl += 32;
        m_mem->pokeW(0x95c5, hl);

}


int JswGame::prevRoom(int room)
{
	int o = room;
	do
	{
		room--;
		if (room < 0) return o;
	} while (roomReserved(room));
	return room;
}

int JswGame::nextRoom(int room)
{
	int o = room;
	do
	{
		room++;
		if (room > 255) return o;
	} while (roomReserved(room));
	return room;
}




JswRoomList *JswGame::getRoomList(void)
{
        int n;
        jswByte *b;
        char rname[40];

        JswRoomList *s = new JswRoomList();

        for (n = 0; n < 256; n++) if (!roomReserved(n))
        {
                b = getRoom(n);
                sprintf(rname, "%03d %-32.32s", n, b + m_roomNameOffset);
                s->addString(rname);
        } 
        return s;
}


int JswGame::canTeleport(void)
{
	if (m_mem->peek(35696) == 205) return 1;
	return 0;	
}



jswByte *JswGame::getTeleportList(void)
{
	int addr;

	if (!canTeleport()) return NULL;

	addr = m_mem->peek(35698) * 256 + m_mem->peek(35697);
	addr = m_mem->peek(addr+6) * 256 + m_mem->peek(addr+5);

	return m_mem->memoryAt(addr);
}

int JswGame::teleUnload(void)
{
	m_mem->poke(35696, 0x01);
	m_mem->poke(35697, 0xFE);
	m_mem->poke(35698, 0xEF);
	return ZXE_CONTINUE;
}

int JswGame::teleLoad(int page)
{
	static jswByte tport[] = {
        	0xD5,			// PUSH DE
		0xDD,0xE5,		// PUSH IX
		0xDD,0x21,0x4A,0x01,	// LD IX,buffer
		0x11,0x04,0x00,		// LD DE,4
		0xDD,0x46,0x00,		// LD B,(IX+0)
		0xDD,0x23,		// INC IX
		0x78,			// LD A,B
       		0xB7,			// OR A
		0x28,0x30,		// JR Z,NOTEL
// LOOP:
		0x3A,0x20,0x84,		// LD A,(SHEET)
		0xDD,0xBE,0x00,		// CP (IX+0)
		0x20,0x24,		// JR NZ,NOTME
		0x3A,0xCF,0x85,		// LD A,(WILLYY)
		0xDD,0xBE,0x01,		// CP (IX+1)
		0x20,0x1C,		// JR NZ,NOTME
		0x3A,0xD3,0x85,		// LD A,(WILLYX)
		0xE6,0x1F,		// AND 1Fh
		0xDD,0xBE,0x02,		// CP (IX+2)
		0x20,0x12,		// JR NZ,NOTME
		0xDD,0x7E,0x03,		// LD A,(IX+3)
		0x32,0x20,0x84,		// LD (SHEET),A
		0xE5,			// PUSH HL
		0x00,0x00,0x00,		// Teleporter FX
		0xE1,			// POP HL
		0xDD,0xE1,		// POP IX
		0xD1,			// POP DE
		0xC1,			// POP BC
		0xC3,0x12,0x89,		// JP L8912
// NOTME:
 		0xDD,0x19,		// ADD IX,DE
		0x10,0xD0,		// DJNZ LOOP
// NOTEL:
		0xDD,0xE1,		// POP IX
		0xD1,			// POP DE
		0x01,0xFE,0xEF,		// LD BC,EFFEh
		0xC9,			// RET
		0x00 };

	memcpy(m_mem->memoryAt(page * 256), tport, 
		sizeof(tport));
	m_mem->poke(page * 256 + 6, page);
        m_mem->poke(35696, 0xCD);
        m_mem->poke(35697, 0x00);
        m_mem->poke(35698, page);
	return ZXE_CONTINUE;
}

int JswGame::teleLoad(void)
{
	int rv;

	JswRoomList *sl = getRoomList();	

	if(!sl) return ZXE_CONTINUE;

	sl->setTitle("Room to be replaced");
	rv = sl->doModal();	

	if (rv >= ZXE_QUIT) { delete sl; return rv; }
	if (rv != ZXE_OK)   { delete sl; return ZXE_CONTINUE; }

	teleLoad(sl->getSelectedRoom() + 0xC0);
	delete sl;

	return ZXE_CONTINUE;
}



int JswGame::toggleTeleporters(void)
{
	int rv;

	if (canTeleport())
	{
		VideoMenu vm("Delete all teleporters?", "Confirm", "Cancel",
				NULL);

		rv = vm.doModal(); 
		if (rv >= ZXE_QUIT) return rv;
		if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;
		teleUnload();
		return rv;	
	}
	return teleLoad();
}



int JswGame::guardianEditor(int room, int guardian, jswByte *table, int x, 
		VideoBitmap *backg)
{
	jswByte theGuard[8];

	memcpy(theGuard, table + 8 * guardian, 8);

	GuardianEditor *g = newGuardianEditor(room, theGuard, backg, x);

	int rv = g->doModal();
	delete g;
	
	if (rv >= ZXE_QUIT) return rv;
	if (rv == ZXE_OK) memcpy(table + 8 * guardian, theGuard, 8);
	return ZXE_CONTINUE;
}

GuardianEditor *JswGame::newGuardianEditor(int room, jswByte *guard, VideoBitmap *b, int x)
{
	return new Jsw128GuardEditor(room, this, guard, b, x);
}


int JswGame::supportsExtGuards(void)
{
	return 0;
}


jswByte *JswGame::getFont(int *len, int *firstc)
{
	int fontPage = canAccent();

	if (fontPage > 0)
	{
		if (len) *len = 32;
		if (firstc) *firstc = 0;
		return m_mem->memoryAt(256 * fontPage);
	}
	if (len) *len = 0; 
	return NULL;
}



int JswGame::clearRooms(void)
{
	jswByte *b;
	VideoMenu vm("This is destructive!", 
                     "Delete all rooms", "Cancel", NULL);

	int rv = vm.doModal();
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;

	for (int n = 0; n < 256; n++)
	{			 
		char buf[33];
		sprintf(buf, "           Room %03d             ",n);
		Room *rc = getRoomClass(n);
		if (rc)
		{
			rc->clear();
			rc->setTitle(buf);
		}
	}
	// Delete all items
	m_mem->poke(m_itemCount, 0xFF);
	return ZXE_CONTINUE;
}


int JswGame::clearSprites(void)
{
       VideoMenu vm("This is destructive!",
                     "Delete all sprites", "Cancel", NULL);

        int rv = vm.doModal();
        if (rv >= ZXE_QUIT) return rv;
        if (rv != ZXE_OK || vm.getSelected() != 0) return ZXE_CONTINUE;

	deleteSprites();
	return ZXE_CONTINUE;
}


int JswGame::clearGuardians(void)
{
	char tbuf[40];
	int gt;
	int rv  = chooseGuardTable(&gt);
	jswByte *b;
	
	if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK)   return ZXE_CONTINUE;

	sprintf(tbuf, "Clear guardian table %04x", gt);
        VideoMenu vm("This is destructive!",
                     tbuf, "Cancel", NULL);
	rv = vm.doModal();
        if (rv >= ZXE_QUIT) return rv;
	if (rv != ZXE_OK)   return ZXE_CONTINUE;

	b = m_mem->memoryAt(gt);
	memset(b, 0, 1023);
	b[1016] = 0xFF;	
	return ZXE_CONTINUE;
}


int JswGame::getItemCount()
{
	return m_mem->peek(m_itemCount);
}

jswByte *JswGame::getTitleAttrs(void)
{
        return m_mem->memoryAt(0x9800);
}        


jswByte JswGame::getWillySpritePage(void) 
{
	return 0x9D;
}


JswGuardian *JswGame::newGuardian(jswByte *guard, int x, int bg)
{
	return new Jsw128Guard(guard, x, m_mem, bg);
}

void JswGame::onChangeMemmap(void)
{
}


typedef Room *RoomPtr;

Room *JswGame::getRoomClass(int room)
{
	Room *r, **rc;
	int newsize, n;

	if (room < 0 || room >= getRoomCount()) return NULL;

	if (room < m_roomCacheSize)
	{
		if (m_roomCache[room]) return m_roomCache[room];
		r = newRoom(room);
		m_roomCache[room] = r;
		return r;
	}

	r = newRoom(room);
	if (m_roomCacheSize == 0) newsize = 64;
	else			  newsize = 2 * m_roomCacheSize;

	// Ensure buffer doesn't overflow!
	while (newsize < room) newsize *= 2;

	rc = new RoomPtr[newsize];
	for (n = 0; n < newsize; n++) rc[n] = NULL;
	if (m_roomCache)
	{
		for (n = 0; n < m_roomCacheSize; n++) rc[n] = m_roomCache[n];
		delete [] m_roomCache;
	}
	m_roomCache = rc;
	m_roomCacheSize = newsize;
	rc[room] = r;
	return r;
}


int JswGame::addGuardians(jswByte *guards, int roomNo, int nguards, int *ids)
{
	return -1;	/* Simple & ineffective */
}

int JswGame::guardiansAvailable(jswByte *guards, int nguards, 
		int tableBase, int count)
{
	int n, m;
	int avail = 0;
	jswByte *table = m_mem->memoryAt(tableBase);
	jswByte *candidate, *guard;

	// Add for each candidate guardian already in the table.
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		for (m = 0; m < nguards; m++)
		{
			guard = guards + 8 * m;
			if (!memcmp(candidate,  guard,  2) &&
			    !memcmp(candidate+3,guard+3,5))
			{
				avail++;
			}
		}
	}
	// Add 1 for each empty slot.
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		if (candidate[0] == 0) avail++;
	}
	// Add 1 for each unused guardian.
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		if (!getRefCount(GUARDIAN(n*8+tableBase))) avail++;
	}
	return avail;
}


int JswGame::addGuardians(jswByte *guards, int roomNo, int nguards, 
		int tableBase, int count, int *ids)
{
	int n, id;

	if (!nguards) return 0;	

	for (n = 0; n < nguards; n++)
	{
		ids[n] = addGuardian(guards + 8 * n, roomNo, tableBase, count);
		addRef(GUARDIAN(tableBase + 8 * ids[n]));
		if (ids[n] < 0)
		{
			for (; n >= 0; n--)
			{
				unRef(GUARDIAN(tableBase + 8 * ids[n]));
			}
			return -1;
		}
	}
	return nguards;
}

int JswGame::addGuardian(jswByte *guard, int roomNo, int tableBase, int count)
{
	int n;
	jswByte *candidate;
	jswByte *table = m_mem->memoryAt(tableBase);

/* First try: See if we can match an existing guard */
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		if (!memcmp(candidate,  guard,  2) &&
		    !memcmp(candidate+3,guard+3,5))
		{
			return n;
		}
	}
/* Second try: See if we can take a blank space */
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		if (candidate[0] == 0)
		{
			memcpy(candidate, guard, 8);
			candidate[2] = 0;
			return n;
		}
	}
/* Third try: See if we can find a guardian with no refs */
	for (n = 0; n < count; n++)
	{
		candidate = table + 8 * n;
		if (!getRefCount(GUARDIAN(n*8+tableBase)))
		{
			memcpy(candidate, guard, 8);
			candidate[2] = 0;
			return n;
		}
	}
/* Nothing doing */
	return -1;	
}


/* Reduce passed sprite to a neat 16x16 square */
int JswGame::addSprite(int fileid, int w, int h, jswByte *bits)
{
	int x, y;
	int usage[2];

	jswByte spriteData[32];
	memset(spriteData, 0, sizeof(spriteData));

	for (y = 0; y < h; y++)
	{
		for (x = 0; x < w; x += 8)
		{
			if (x >= 32) break;
			spriteData[(y * 16 + x)/8] = bits[(y * w + x)/ 8];
		}
		if (y >= 32) break;
	}
	usage[0] = -2;
	usage[1] = -3;
	return addSingleSprite(fileid, spriteData, usage);
}


/* We have a sprite and its file ID. See where in memory it might be put. */
int JswGame::addSingleSprite(int id, jswByte *bits, int roomids[2])
{
	int n, max, m;
	jswByte *c;
	jswByte bits2[256];
	SpriteList list(getMem(), 0, 0, 50, 50);
	fillSprites(-1, &list, true);

#if DEBUG_SPRITE_IMPORT
	printf("addSingleSprite:id=%x\n", id);
#endif
	max = list.getItemCount();
	// First, see if the sprite is already present in any of the existing
	// pages.
	for (n = 0; n < max; n++)
	{
		SpriteEntry e = list.getItem(n);
#if DEBUG_SPRITE_IMPORT
		printf("Check for re-use: %04x\n", e.zxAddress);
#endif
		c = m_mem->memoryAt(e.zxAddress);
//		printf("%02x %02x %02x %02x vs %02x %02x %02x %02x\n",
//				c[0], c[1], c[2], c[3],
//				bits[0], bits[1], bits[2], bits[3]);

		if (!memcmp(c, bits, 32)) 
		{
#if DEBUG_SPRITE_IMPORT
			printf("Re-using %04x\n", e.zxAddress);
#endif
			addRef(SPRITE(e.zxAddress));
			return e.zxAddress;
		}
	}
	// JSW64: Check for local sprites
	if (roomids[1] == -1)	// Only used in one room
	{
		n = getRoomClass(roomids[0])->addLocalSprite(bits);
		if (n >= 0) return n;
	}
	// Next, see if there's an unreferenced sprite available in any of
	// the pages.
	for (n = 0; n < max; n++)
	{
		SpriteEntry e = list.getItem(n);

#if DEBUG_SPRITE_IMPORT
		printf("Scanning sprite entry %x\n", e.zxAddress);
#endif
		if (getRefCount(SPRITE(e.zxAddress)))
		{
#if DEBUG_SPRITE_IMPORT
			printf("-- used\n");
#endif
			continue;
		}
#if DEBUG_SPRITE_IMPORT
		printf("Inserting at %x\n", e.zxAddress);
#endif
		c = m_mem->memoryAt(e.zxAddress);
		memcpy(c, bits, 32);
		addRef(SPRITE(e.zxAddress));
		return e.zxAddress;
	}
	// That didn't work. We'll have to allocate a whole new page.
	memset(bits2, 0, sizeof(bits2));
	memcpy(bits2, bits, 32);
#if DEBUG_SPRITE_IMPORT
	printf("Creating new sprite page\n");
#endif
	return addSpritePage(bits2, roomids);
}



int JswGame::addSpritePage(jswByte *bits, int roomids[2])
{
	int n, max, m, used;
	jswByte *c;
	SpriteList list(getMem(), 0, 0, 50, 50);
	fillSprites(-1, &list, false);

	max = list.getItemCount();
	// First, see if the sprite page is already present
	for (n = 0; n < max; n++)
	{
		SpriteEntry e = list.getItem(n);
#if DEBUG_SPRITE_IMPORT
		printf("Check for re-use: %04x\n", e.zxAddress);
#endif
		c = m_mem->memoryAt(e.zxAddress);
	//	printf("%02x %02x %02x %02x vs %02x %02x %02x %02x\n",
	//			c[0], c[1], c[2], c[3],
	//			bits[0], bits[1], bits[2], bits[3]);

		if (!memcmp(c, bits, 256)) 
		{
#if DEBUG_SPRITE_IMPORT
			printf("Re-using %04x\n", e.zxAddress);
#endif
			addRef(SPRITE(e.zxAddress));
			return e.zxAddress;
		}
	}
	// JSW64/MM: Check for local sprites here. 
	if (roomids[1] == -1)	// Only used in one room
	{
		n = getRoomClass(roomids[0])->addLocalSpritePage(bits);
		if (n >= 0) return n;
	}

	// Next, see if there's an unreferenced sprite page.
	for (n = 0; n < max; n++)
	{
		SpriteEntry e = list.getItem(n);

		used = 0;
		for (m = 0; m < 8; m++)
		{
			if (getRefCount(SPRITE(e.zxAddress + 32 * m)))
			{
				used = 1;
				break;
			}
		}
	
		if (!used)
		{
			c = m_mem->memoryAt(e.zxAddress);
			memcpy(c, bits, 256);
			addRef(SPRITE(e.zxAddress));
			return e.zxAddress;
		}
	}
	return -1;
}

void JswGame::clearPageMap(int nsprites)
{
	int x, y;

	if (m_isp) delete [] m_isp;
	m_isp = new ImportedSpritePage[nsprites];
	m_spriteCount = nsprites;
	for (y = 0; y < nsprites; y++)
	{
		m_isp[y].usedInRooms[0] = -1;
		m_isp[y].usedInRooms[1] = -1;
		memset(m_isp[y].bits, 0, sizeof(m_isp[y].bits));
		for (x = 0; x < 8; x++) m_isp[y].spriteIDs[x] = -1;
	}
}

void JswGame::dumpPageMap()
{
	int y, x, some;

	printf("pageMap:\n");
	for (y = 0; y < m_spriteCount; y++)
	{
		int used = 0;
		for (x = 0; x < 8; x++) if (m_isp[y].spriteIDs[x] != -1) 
			used = 1;
		if (!used) continue;
		printf("%3d. ", y);
		for (x = 0; x < 8; x++)
		{
			printf("%06x ", m_isp[y].spriteIDs[x] & 0xFFFFFF);
		}
		printf("(%d %d)", m_isp[y].usedInRooms[0], 
				m_isp[y].usedInRooms[1]);
		printf("\n");
	}
}


void JswGame::addPageMap(int room, ImportedSprite pageSet[8])
{
	int x, y, found;

//	printf("Add page map: ");
	found = 0;
	for (x = 0; x < 8; x++)
	{
//		printf("%06x ", pageSet[x].spriteID & 0xFFFFFF);
		if (pageSet[x].spriteID != -1) found = 1;
	}
//	printf("\n");
	if (!found) return;	/* Nothing to add */

/* We have pageSet, a list of up to 8 sprites. This has to be merged with 
 * pageMap.
 * 
 * There are two possibilities:
 * 1. One or more of the IDs in the sprite list is already present in one
 *   of the sprite pages in memory. If so, copy these sprites into it. 
 */
	found = 0;
	for (y = 0; y < m_spriteCount; y++)
	{
		found = 1;
		for (x = 0; x < 8; x++)
		{
			if (pageSet[x].spriteID == -1) continue;
			if (pageSet[x].spriteID != m_isp[y].spriteIDs[x]) 
				found = 0;
		}
		if (found) break;
	}
	if (found) 
	{
		for (x = 0; x < 8; x++) 
		{
			if (pageSet[x].spriteID == -1) continue;
			if (pageSet[x].spriteID != m_isp[y].spriteIDs[x] &&
			    m_isp[y].spriteIDs[x] != -1)
			{
				fprintf(stderr, 
"Internal error or bad file format: Two sprites mapped in the same position!\n"
"IDs are %x and %x\n", pageSet[x].spriteID, m_isp[y].spriteIDs[x]);
				continue;
			}
			m_isp[y].spriteIDs[x] = pageSet[x].spriteID;
			memcpy(m_isp[y].bits + 32 * x, pageSet[x].bits, 32); 
		}
		if (m_isp[y].usedInRooms[0] == -1) 
			m_isp[y].usedInRooms[0] = room;
		else if (m_isp[y].usedInRooms[1] == -1)
			m_isp[y].usedInRooms[1] = room;
		return;	
	}

/* 2. The sprite ID is not known. Give it a new page. */

	found = 0;
	for (y = 0; y < m_spriteCount; y++)
	{
		found = 1;
		for (x = 0; x < 8; x++)
		{
			if (m_isp[y].spriteIDs[x] != -1) 
			{
				found = 0;
				break;
			}
		}
		if (found) break;
	}
	if (!found) 
	{
		fprintf(stderr, "Internal error: Sprite table overflow\n");
		return;
	}
	/* y = index of new sprite page */
	for (x = 0; x < 8; x++) 
	{
		m_isp[y].spriteIDs[x] = pageSet[x].spriteID;
		memcpy(m_isp[y].bits + 32 * x, pageSet[x].bits, 32); 
	}
	m_isp[y].usedInRooms[0] = room;
}



void JswGame::addPageMap(int room, ImportedSprite is)
{
	int x, y, found;

/* See if this page is already mapped */
	found = 0;
	for (y = 0; y < m_spriteCount; y++)
	{
		for (x = 0; x < 8; x++)
		{
			if (m_isp[y].spriteIDs[x] == is.spriteID)
			{
				if (m_isp[y].usedInRooms[0] == -1) 
					m_isp[y].usedInRooms[0] = room;
				else if (m_isp[y].usedInRooms[1] == -1)
					m_isp[y].usedInRooms[1] = room;
				return;
			}
		}
	}

/* No. Add it. */
	for (y = 0; y < m_spriteCount; y++)
	{
		found = 1;
		for (x = 0; x < 8; x++)
		{
			if (m_isp[y].spriteIDs[x] != -1) 
			{
				found = 0;
				break;
			}
		}
		if (found) break;
	}
	if (!found)
	{
		fprintf(stderr, "Internal error: Sprite table overflow\n");
		return;
	}
	m_isp[y].spriteIDs[0] = is.spriteID;
	memcpy(m_isp[y].bits, is.bits, 32);
	m_isp[y].usedInRooms[0] = room;
}



void JswGame::compactPageMap()
{
	int x, y, y1;
	int used[2];

	for (y = 0; y < m_spriteCount; y++)
	{
		used[0] = 0;
		for (x = 0; x < 8; x++) 
		{
			if (m_isp[y].spriteIDs[x] >= 0) used[0] |= (1 << x); 
		}	
		// Can't merge empty or full lines 
		if (used[0] == 0 || used[0] == 0xFF) continue;
		for (y1 = y + 1; y1 < m_spriteCount; y1++)
		{
			used[1] = 0;
			for (x = 0; x < 8; x++) 
			{
				if (m_isp[y1].spriteIDs[x] >= 0) 
					used[1] |= (1 << x); 
			}	
			// Can't merge empty or full lines 
			if (used[1] == 0 || used[1] == 0xFF) 
				continue;
			if (used[0] & used[1]) continue; // Lines overlap

// Two part-used lines - a match made in heaven 
			for (x = 0; x < 8; x++)
			{
				if (m_isp[y].spriteIDs[x] == -1) 
				{
					m_isp[y].spriteIDs[x] = m_isp[y1].spriteIDs[x];
					memcpy(m_isp[y]. bits + 32 * x,
					       m_isp[y1].bits + 32 * x, 32);
// Merge the lists of rooms:
// If the first entry has no room refs at all, copy in both refs from the 
// second entry. 
					if (m_isp[y].usedInRooms[0] == -1)
					{
						m_isp[y].usedInRooms[0] = m_isp[y1].usedInRooms[0];
						m_isp[y].usedInRooms[1] = m_isp[y1].usedInRooms[1];
					}
// If the first entry has one used and one free, and the second has an entry
// that the first one doesn't, copy that in.
					else if (m_isp[y].usedInRooms[1] == -1)
					{
						int n;
						for (n = 0; n < 2; n++)
						{
							if (m_isp[y1].usedInRooms[n] >= 0 && m_isp[y1].usedInRooms[n] != m_isp[y].usedInRooms[0])
							{
								m_isp[y].usedInRooms[1] = m_isp[y1].usedInRooms[n];
								break;
							}
						}
					}
					m_isp[y1].spriteIDs[x] = -1;
				}
			}
			used[0] |= used[1];
			if (used[0] == 0xFF) break; // Page full
		}
	}
}


bool JswGame::loadSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent)
{
        xmlNodePtr cur;
	int memid, fileid, w, h, x, xpos, hexv;
	char *t, *t1, hex[3];
	int count = 0, nspr;

	/* Make a count of valid <sprite> nodes */
	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
		if (!s) continue;
		xmlFree(s);
		s = (char *)xmlGetProp(cur, (xmlChar *)"w");
		if (!s) continue;
		w = atoi(s);
		xmlFree(s);
		if (!w) continue;	
		s = (char *)xmlGetProp(cur, (xmlChar *)"h");
		if (!s) continue;
		h = atoi(s);
		xmlFree(s);
		if (!h) continue;	
		t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if (!t) continue;
		xmlFree(t);
		++count;
	}
//	fprintf(stderr, "There are %d valid sprite nodes\n", count);
	if (!count) return false;

	if (m_sprites) delete [] m_sprites;
	if (m_isp)     delete [] m_isp;
	m_isp = NULL;
	m_sprites = new ImportedSprite[count];
	m_spriteCount = count;
	clearPageMap(count);

	nspr = 0;
	for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
	{
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
		if (!s) continue;
		fileid = 0;
		sscanf(s, "%x", &fileid);
		xmlFree(s);
		s = (char *)xmlGetProp(cur, (xmlChar *)"w");
		if (!s) continue;
		w = atoi(s);
		xmlFree(s);
		if (!w) continue;	
		s = (char *)xmlGetProp(cur, (xmlChar *)"h");
		if (!s) continue;
		h = atoi(s);
		xmlFree(s);
		if (!h) continue;	
		t = (char *)xmlNodeListGetString(doc, cur->xmlChildrenNode, 1);
		if (!t) continue;

		assert(nspr < m_spriteCount);
		m_sprites[nspr].spriteID = fileid;
		t1 = t;
		x = 0;
		hex[0] = 0;
		xpos = 0;
		while (t1[0])
		{
			if (!isxdigit(t1[0]))
			{
				++t1;
				continue;
			}
			if (hex[0] == 0) 
			{
				hex[0] = *t1;
				hex[1] = 0;
				++t1;
				continue;
			}
			hex[1] = *t1;
			hex[2] = 0;
			++t1;
			hexv = 0;
			sscanf(hex, "%x", &hexv);
			hex[0] = 0;
			assert(nspr < m_spriteCount);
			m_sprites[nspr].bits[x++] = hexv;
			xpos += 8;
			if (xpos >= 16) 
			{
/* Skip any pixels to the right of column 16 */
				for (; xpos < w; xpos += 8) 
				{
					while (!isxdigit(t1[0])) ++t1;
					++t1;
					while (!isxdigit(t1[0])) ++t1;
					++t1;
				}
			}
			if (x >= 32) break;
		}
		++nspr;
		xmlFree(t);
	}
	return true;
}


ImportedSprite JswGame::findSprite(int xref)
{
	int n;
	ImportedSprite nil;

	for (n = 0; n < m_spriteCount; n++)
	{
		if (m_sprites[n].spriteID == xref) return m_sprites[n];
	}
	nil.spriteID = -1;
	return nil;
}


// Import sprites from file. This completely replaces the version in Game,
// because of the way JSW groups sprites into pages.
bool JswGame::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int *roomMap)
{
// Sprites are stored in the XML file as individual images. We have to 
// combine them back into pages. This is done by walking the file looking
// for <spriteref> nodes, which define groupings of sprites. 
//
	SpriteList list(getMem(), 0, 0, 50, 50);
	xmlNodePtr root, room, guardians, guardian, spriteref, cur;
	char *sframe, *sxref, *sg1, *sg2;
     	int g1, g2, frame, xref, x, y, y1;
	ImportedSprite pageSet[8];
	int n, m, found, match, count, roomNumber;

	root = xmlDocGetRootElement(doc);
	if (!root) return false;
	if (!loadSprites(doc, ns, parent)) return true;	// Nothing to load

// Handle all nodes of form rooms/room/guardians/??/spriteref 
// and rooms/room/hints/hint/spriteref 
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

// Only process rooms which are in the map (& hence will be imported) 
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		xmlFree(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) 
				{
					roomNumber = n;
					found = 1; 
				}
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
//		printf("Importing sprites for room %d\n", roomNumber);
		getRoomClass(roomNumber)->prepareForSpriteImport();
// Room found. Now scan its spriterefs 
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
// Start with any in hints (these are for Willy's sprite)
			if (!strcmp((char *)guardians->name, "hints"))
			{
				for (guardian = guardians->xmlChildrenNode;
				     guardian;
				     guardian = guardian->next)
			    {
				if (guardian->ns != ns) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"type");
				if (sg1 && !strcmp(sg1, "sprite"))
				{
				  xmlFree(sg1);
				  sg1 = NULL;
				  for (x = 0; x < 8; x++) 
					  pageSet[x].spriteID = -1;
				  for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				  {
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
					{
						xmlFree(sxref);
						continue;
					}
					xmlFree(sxref);
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					if (!sframe) continue;
					frame = atoi(sframe);
					xmlFree(sframe);
					pageSet[frame] = findSprite(xref);
				  }
				  addPageMap(roomNumber, pageSet);
				}
				else if (sg1) xmlFree(sg1);
			    }
			    continue;
			}

			if (strcmp((char *)guardians->name, "guardians")) 
				continue;
			for (guardian = guardians->xmlChildrenNode;
				guardian;
				guardian = guardian->next)
			{
				if (guardian->ns != ns) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"framemask");
				sg2 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"startframe");
				if (sg1) g1 = atoi(sg1); else g1 = 0;
				if (sg2) g2 = atoi(sg2); else g2 = 0;
				if (sg1) xmlFree(sg1);
				if (sg2) xmlFree(sg2);
				if (!strcmp((char *)(guardian->name), 
							"skylab"))
				{
					g1 = 7;
				}
				for (x = 0; x < 8; x++) 
					pageSet[x].spriteID = -1;
				for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
					{
						xmlFree(sxref);
						continue;
					}
					xmlFree(sxref);
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					if (!sframe) continue;
					frame = atoi(sframe);
					xmlFree(sframe);
					pageSet[((frame & g1) | g2)] = findSprite(xref);
				}
				addPageMap(roomNumber, pageSet);
			}
		}
	}
// Dump what we have so far. 
#if DEBUG_SPRITE_IMPORT
	dumpPageMap();
#endif

/* Now see if any of the sprite pages we have so far match existing 
 * sprites. */
	for (n = 0; n < m_spriteCount; n++)
	{
		fillSprites(-1, &list, false);
		int max = list.getItemCount();
		for (m = 0; m < max; m++)
		{
			SpriteEntry e = list.getItem(m);
			int match = 1, some = 0;
			for (x = 0; x < 8; x++)
			{
				if (m_isp[n].spriteIDs[x] < 0) continue;
				// xref = address of memory sprite
				some = 1;
				xref = e.zxAddress + 0x20 * x;
#if DEBUG_SPRITE_IMPORT
printf("Comparing sprites: %04x with %04x\n", m_isp[n].spriteIDs[x], xref);
#endif
				jswByte *src = m_mem->memoryAt(xref);
				if (memcmp(src, m_isp[n].bits + 32 * x, 32))
				{
					match = 0;
					break;
				}
			}
			if (match && some) 
			{
				// Report a match.
#if DEBUG_SPRITE_IMPORT
				printf("Matched row: ");
				for (x = 0; x < 8; x++) printf("%06x ", 
						m_isp[n].spriteIDs[x]);
				printf("with sprite page %04x\n",
						e.zxAddress);
#endif
/* Add mappings for the sprites, addRef to them so they don't get overwritten,
 * and remove them from the pending list */
				for (x = 0; x < 8; x++)
				{
					if (m_isp[n].spriteIDs[x] < 0) continue;
					xref = e.zxAddress + 0x20 * x;
					addSpriteRef(m_isp[n].spriteIDs[x],xref);
					addRef(SPRITE(xref));
					m_isp[n].spriteIDs[x] = -1;
				}
				break;	// out of reference count scan 
			}
		}
	}
// Now compact the map, so that any pair of sprites that can share a page
// will do so.
	compactPageMap();
// Now add those sprite pages that are left over.
	for (y = 0; y < m_spriteCount; y++)
	{
		int used = 0;
		for (x = 0; x < 8; x++) if (m_isp[y].spriteIDs[x] != -1) 
			used = 1;
		if (!used) continue;
		int memid = addSpritePage(m_isp[y].bits, m_isp[y].usedInRooms);
		if (memid < 0) continue;
		for (x = 0; x < 8; x++)
		{
			addSpriteRef(m_isp[y].spriteIDs[x],memid + 32 * x);
			addRef(SPRITE(memid + 32 * x));
			m_isp[y].spriteIDs[x] = -1;
		}
	}	
// Dump what we have left over, which should be nothing.
#if DEBUG_SPRITE_IMPORT
	dumpPageMap();
#endif
/* That has done all sprites that come in groups; pageMap should by now be 
 * empty. Next step is to put all the portal sprites in the page map, each
 * in its own page. The simplistic code below brings in every sprite in the 
 * file, but we don't want to do that; instead, just bring in portal sprites 
 * as they're the only ones that aren't grouped into pages. 
        for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
        {
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
                if (!s) continue;
                xref = 0;
                sscanf(s, "%x", &xref);
		addPageMap(xref);	
	} */
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

/* Only process rooms which are in the map (& hence will be imported) */
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		xmlFree(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) found = 1; 
				roomNumber = n;
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
			if (!strcmp((char *)guardians->name, "portal"))
			{
				char *s = (char *)xmlGetProp(guardians, 
						(xmlChar *)"sprite");
                		if (!s) continue;
                		xref = 0;
                		sscanf(s, "%x", &xref);
				xmlFree(s);
				pageSet[0] = findSprite(xref);
				addPageMap(roomNumber, pageSet[0]);
			}
		}
	} 
// Dump what we loaded.
#if DEBUG_SPRITE_IMPORT
	dumpPageMap();
#endif
	for (y = 0; y < m_spriteCount; y++)
	{
		if (m_isp[y].spriteIDs[0] >= 0)
		{
			int memid = addSingleSprite(m_isp[y].spriteIDs[0],
					m_isp[y].bits,
					m_isp[y].usedInRooms);
			if (memid >= 0)
			{
				addSpriteRef(m_isp[y].spriteIDs[0],memid);
				addRef(SPRITE(memid));
			}
			m_isp[y].spriteIDs[0] = -1;
		}
	}
// Dump what's left, which should be nothing.
#if DEBUG_SPRITE_IMPORT
	dumpPageMap();
#endif
/* Simplistic call to the base class version
	int r = Game::importSprites(doc, ns, parent);
	compactSprites();
	return r;
*/
	compactSprites();
	return true;
}



#if 0
bool JswGame::importSprites(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr parent,
		int *roomMap)
{
// Sprites are stored in the XML file as individual images. We have to 
// combine them back into pages. This is done by walking the file looking
// for <spriteref> nodes, which define groupings of sprites. 
//
	xmlNodePtr root, room, guardians, guardian, spriteref, cur;
	char *sframe, *sxref, *sg1, *sg2;
     	int g1, g2, frame, xref, x, y, y1;
	int pageSet[8];
	int n, found, match;

	root = xmlDocGetRootElement(doc);
	if (!root) return false;
	clearPageMap();

/* Handle all nodes of form rooms/room/guardians/??/spriteref */
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

/* Only process rooms which are in the map (& hence will be imported) */
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) found = 1; 
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
/* Also rooms/room/hints/hint/spriteref */
			if (!strcmp((char *)guardians->name, "hints"))
			{
				for (guardian = guardians->xmlChildrenNode;
				     guardian;
				     guardian = guardian->next)
			    {
				if (guardian->ns != ns) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"type");
				if (sg1 && !strcmp(sg1, "sprite"))
				{
				  for (x = 0; x < 8; x++) pageSet[x] = -1;
				  for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				  {
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sframe) continue;
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
						continue;
					frame = atoi(sframe);

					pageSet[frame] = xref;
				  }
				  addPageMap(pageSet);
				}
			    }
			    continue;
			}

			if (strcmp((char *)guardians->name, "guardians")) 
				continue;
			for (guardian = guardians->xmlChildrenNode;
				guardian;
				guardian = guardian->next)
			{
				if (guardian->ns != ns) continue;
				sg1 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"framemask");
				sg2 = (char *)xmlGetProp(guardian, 
						(xmlChar *)"startframe");
				if (sg1) g1 = atoi(sg1); else g1 = 0;
				if (sg2) g2 = atoi(sg2); else g2 = 0;

				for (x = 0; x < 8; x++) pageSet[x] = -1;
				for (spriteref = guardian->xmlChildrenNode;
					spriteref;
					spriteref = spriteref->next)
				{
					if (spriteref->ns != ns) continue;
					if (strcmp((char *)spriteref->name, 
						"spriteref")) continue;
					sframe = (char *)xmlGetProp(
						spriteref, (xmlChar *)"frame");
					sxref = (char *)xmlGetProp(
						spriteref, (xmlChar *)"xref");
					if (!sframe) continue;
					if (!sxref) continue;
					if (!sscanf(sxref, "%x", &xref))
						continue;
					frame = atoi(sframe);

					pageSet[((frame & g1) | g2)] = xref;
				}
				addPageMap(pageSet);
			}
		}
	}
/* Now compact the map. */
	compactPageMap();
/* Now add in sprites that don't come in groups. The simplistic code below
 * brings in every sprite in the file, but we don't want to do that; 
 * instead, just bring in portal sprites as they're the only ones that
 * aren't grouped into pages. 
        for (cur = parent->xmlChildrenNode; cur; cur = cur->next)
        {
		if (cur->ns != ns) continue;
		if (strcmp((char *)cur->name, "sprite")) continue;
		char *s = (char *)xmlGetProp(cur, (xmlChar *)"base");
                if (!s) continue;
                xref = 0;
                sscanf(s, "%x", &xref);
		addPageMap(xref);	
	} */
	for (room = root->xmlChildrenNode; room; room = room->next)
	{
		if (room->ns != ns) continue;
		if (strcmp((char *)room->name, "room")) continue;

/* Only process rooms which are in the map (& hence will be imported) */
		char *num = (char *)xmlGetProp(room, (xmlChar *)"number");
		if (!num) continue;
		found = 0;
		match = atoi(num);
		for (n = 0; n < 256; n++)
		{
			if (roomMap[n] == match) 
			{ 
		/* Does mapped room actually exist, though? */
				if (getRoom(n)) found = 1; 
				break; 
			}
		}
		if (!found) 
		{
//			printf("Room %d in file not mapped in database\n",
//					match);
			continue;
		}
		for (guardians = room->xmlChildrenNode; guardians; 
			       guardians = guardians->next)	
		{
			if (guardians->ns != ns) continue;
			if (!strcmp((char *)guardians->name, "portal"))
			{
				char *s = (char *)xmlGetProp(guardians, 
						(xmlChar *)"sprite");
                		if (!s) continue;
                		xref = 0;
                		sscanf(s, "%x", &xref);
				addPageMap(xref);	
			}
		}
	} 

/* Now we know where every sprite is going. All we have to do, is actually
 * bring them in. */
/**/
printf("pageMap:\n");
for (y = 0; y < m_mapMax; y++)
{
printf("%3d.[%02x] ", y, m_mapUsed[y]);
for (x = 0; x < 8; x++)
{
	printf("%05x ", m_pageMap[y][x] & 0xFFFFF);
}
printf("\n");
}


	int r = Game::importSprites(doc, ns, parent);
	compactSprites();
	return r;
}
#endif // 0


void JswGame::addFixedRefs()
{
	int n;

	/* Add refs for Willy, the barrel, Maria and the toilet, so they aren't
	 * overwritten by an import */
        int adrw = 256 * getWillySpritePage();
        int adrm = 256 * getMem()->peek(0x955D) + 0x80;
        int adrt = 256 * getMem()->peek(0x95B3);
        int adrb = 256 * getMem()->peek(0x8c68) + getMem()->peek(0x8c67);
        int adrf = 256 * getMem()->peek(0x8c82) + getMem()->peek(0x8c81);


	addRef(SPRITE(adrb)); // Barrel
	addRef(SPRITE(adrf)); // Foot

	for (n = 0; n < 0x80;  n += 0x20) addRef(SPRITE(adrm + n));
	for (n = 0; n < 0x100; n += 0x20) addRef(SPRITE(adrw + n));
	for (n = 0; n < 0x80;  n += 0x20) addRef(SPRITE(adrt + n));
}


// Merge identical sprite pages. This ought not to be necessary because the
// compaction should have happend during load. But do it anyway.
void JswGame::compactSprites()
{
	int n, max, m, x, y;
	jswByte *c1, *c2;
	SpriteList list(getMem(), 0, 0, 50, 50);
	fillSprites(-1, &list, false);

	max = list.getItemCount();
	// First, see if the sprite is already present
	for (n = 0; n < max; n++)
	{
		for (m = 0; m < n; m++)
		{
			SpriteEntry e1 = list.getItem(m);
			SpriteEntry e2 = list.getItem(n);
			c1 = m_mem->memoryAt(e1.zxAddress);
			c2 = m_mem->memoryAt(e2.zxAddress);
			if (!memcmp(c1, c2, 256)) 
			{
//				printf("Merging sprite pages %x and %x\n",
//						e1.zxAddress, e2.zxAddress);
// Divert all c2 references to c1. Firstly in the game sprite map
				for (x = 0; x < m_spriteMapLen; x += 2)
				{
					if (m_spriteMap[x] == e2.zxAddress)
					{
						m_spriteMap[x + 1] = e1.zxAddress;
					}
				}
// And then in the reference counts.
				for (y = 0; y < 256; y += 32)
				{
					x = getRefCount(SPRITE(e2.zxAddress+y));
					while (x--) 
					{
						addRef(SPRITE(e1.zxAddress+y));
						unRef(SPRITE(e2.zxAddress+y));
					}
				}
// Free the sprite page 
				freeSpritePage(e2.zxAddress);
			}
		}
	}
}


void JswGame::freeSpritePage(int address)
{

}

void JswGame::checkMariaToilet()
{
	if (getMariaRoom() == getToiletRoom())
	{
		m_mem->poke(0x9539, 0x18);
		m_mem->poke(0x9567, 0xB6);
		m_mem->poke(0x959A, 0xC0);
		m_mem->poke(0x959B, 0xCD);
		m_mem->poke(0x959C, 0x3B);
		m_mem->poke(0x959D, 0x95);
		m_mem->poke(0x959E, 0x00);
		m_mem->poke(0x959F, 0x00);
	}
	else
	{
		m_mem->poke(0x9539, 0x20);
		m_mem->poke(0x9567, 0xB7);
		m_mem->poke(0x959A, 0x3A);
		m_mem->poke(0x959B, 0x20);
		m_mem->poke(0x959C, 0x84);
		m_mem->poke(0x959D, 0xFE);
		m_mem->poke(0x959E, getToiletRoom());
		m_mem->poke(0x959F, 0xC0);
	}
}


int JswGame::getDisplayBorder(int room)
{
	return getRoomClass(room)->getBorder();
}


jswByte *JswGame::getRoomTitleAttrs()
{
	return getTitleAttrs() + 0x200;
}

