/************************************************************************

    JSWED 2.3.0 - Editor for Jet Set Willy and derivatives

    Copyright (C) 2005  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "jswed.hxx"
#include "tune128.hxx"
#include <assert.h>

#define CHUNK 128

static NoteSequence *glav = NULL;
static int glcnt = 0;

NoteSequence::NoteSequence()
{
	m_max = 0;
	m_count = 0;
	m_data = NULL;
}


NoteSequence::NoteSequence(int len)
{
	m_max = len;
	m_count = len;
	m_data = new Note128[m_max];
}

NoteSequence::~NoteSequence()
{
	if (m_data)
	{
		delete [] m_data;
	}
}

void NoteSequence::assertValid()
{
	assert(m_count <= m_max);
	if (!m_data) assert(m_count == 0);
}

void NoteSequence::append(Note128 datum)
{
	assertValid();
	if (m_count >= m_max)
	{
		if (!m_data)
		{
			assertValid();
			m_max = CHUNK;
			assertValid();
			glav = this; glcnt = 0;
			Note128 *d2 = new Note128[m_max];
			glav = NULL;
			assertValid();
			m_data = d2;
			assertValid();
		}
		else
		{
			assert(m_max != 0);
			assert(2 * m_max > m_count);
			assertValid();
			glav = this; glcnt = 0;
			Note128 *d2 = new Note128[2 * m_max];
			glav = NULL;
			for (int n = 0; n < m_count; n++)
			{
				d2[n].copy(&m_data[n]);
			}
			assertValid();
			glav = this; glcnt = 0;
			delete [] m_data;
			glav = NULL;
			m_data = d2;
			assertValid();
			m_max *= 2;
		}
	}
	assertValid();
	m_data[m_count++].copy(&datum);
	assertValid();
}

void NoteSequence::append(const NoteSequence &other)
{
	assertValid();
	for (int n = 0; n < other.length(); n++)
	{
		append(other.m_data[n]);
	}
	assertValid();
}

void NoteSequence::dump(FILE *fp)
{
	for (int n = 0; n < m_count; n++) m_data[n].dump(fp);
}

void NoteSequence::subset(NoteSequence *s, int first, int count)
{
	s->clear();
	if (count == -1) count = m_count - first;
	for (int n = 0; n < count; n++)
	{
		if (n + first >= m_count) break;
		s->append(m_data[n + first]);
	}
}


bool NoteSequence::equals(NoteSequence *other)
{
	if (m_count != other->m_count) return false;
	for (int n = 0; n < m_count; n++)
		if (!(m_data[n] == other->m_data[n]))
			return false;
	return true;
}



bool NoteSequence::operator ==(NoteSequence &other)
{
	return equals(&other);
}



/* Number of bytes this string of notes would occupy */
unsigned NoteSequence::memLength()
{
	int l = 0;

	for (int n = 0; n < m_count; n++) l += m_data[n].memLength();
	return l;
}


unsigned NoteSequence::store(SpectrumMemory *mem,
		                unsigned base, unsigned addr)
{
	unsigned addr0 = addr;

	for (int n = 0; n < m_count; n++)
	{
		addr += m_data[n].store(mem, base, addr);
	}
	return addr - addr0;
}



Note128::Note128()
{
	if (glav)
	{
		glav->assertValid();
	}
	m_body = NULL;
	clear();
	if (glav)
	{
		glav->assertValid();
		++glcnt;
	}
}


Note128::Note128(unsigned type, unsigned p1)
{
	if (glav)
	{
		glav->assertValid();
	}
	m_body = NULL;
	clear();
	m_type = type;
	switch(m_type)
	{
		case TUNEDJNZ: m_repeats = p1;
			       m_body = new NoteSequence();
			       break;
		case TUNERET:  break;
		case TUNECALL: m_offset = p1; break;
		case TUNESTOP: break;
		default: if (isNote()) m_duration = p1;
			 else          m_effect = p1;
	}
	if (glav)
	{
		glav->assertValid();
		++glcnt;
	}
}

Note128::Note128(NoteSequence &s)
{
	if (glav)
	{
		glav->assertValid();
	}
	clear();
	m_body = new NoteSequence();
	m_type = TUNEDJNZ;
	m_body->append(s);
	m_repeats = 1;
	if (glav)
	{
		glav->assertValid();
		++glcnt;
	}
}


Note128::~Note128()
{
	if (m_body) delete m_body;
}

void Note128::clear()
{
	m_type = 0;
	m_duration = 0;
	m_effect = 0;
	m_repeats = 0;
	m_offset = 0;
}

bool Note128::operator == (Note128 &other)
{
	if (m_type != other.m_type) return false;
	switch(m_type)
	{
		case TUNEDJNZ: if (m_repeats != other.m_repeats) return false;
			       return (m_body->equals(other.m_body));
		case TUNERET:  return true;
		case TUNECALL: return (m_offset == other.m_offset);
		case TUNESTOP: return true;
		default: if (isNote()) 
			      return (m_duration == m_duration);
			 else return (m_effect   == m_effect);
	}
}



void Note128::copy(Note128 *other)
{
	int n;

	if (m_body) delete m_body;
	m_body = NULL;
	clear();
	m_type = other->m_type;
	switch(m_type)
	{
		case TUNEDJNZ:
			n = other->m_body->length();
			m_body = new NoteSequence(n);
			m_body->append(*other->m_body);	
			m_repeats = other->m_repeats;
			break;	
		case TUNERET:  break;
		case TUNECALL: m_offset = other->m_offset; break;
		case TUNESTOP: break;
		default: if (isNote()) m_duration = other->m_duration;
			 else          m_effect   = other->m_effect;
			 break;
	}
}


unsigned Note128::memLength(void)
{
	switch(m_type)
	{
		case TUNEDJNZ: return 4 + m_body->memLength();
		case TUNERET:  return 1;
		case TUNECALL: return 3;
		case TUNESTOP: return 1;
		default:       return 2;
	}
}

void Note128::dump(FILE *fp)
{
	switch(m_type)
	{
		case TUNEDJNZ: fprintf(fp, "LOOP (%u) [", m_repeats);
			       m_body->dump(fp);
			       fprintf(fp, "];");
			       break;
		case TUNERET:  fprintf(fp, "RET ");
			       break;
		case TUNECALL: fprintf(fp, "CALL (%u);", m_offset);
		case TUNESTOP: fprintf(fp, "STOP ");
			       break;
		case 0:	       fprintf(fp, "REST (%u);", m_duration);
			       break;
		default:       if (isNote()) fprintf(fp, "TONE (%u,%u);",
					       m_type, m_duration);
				else	     fprintf(fp, "FX (%u,%u);",
						m_type, m_effect);
				break;
	}
}

unsigned Note128::store(SpectrumMemory *mem, unsigned base, unsigned addr)
{
	unsigned offset;

	switch(m_type)
	{
		case TUNEDJNZ:
			offset = m_body->store(mem, base, addr);
			addr += offset;
			if (mem)
			{
				mem->poke(addr,     TUNEDJNZ);
				mem->poke(addr + 1, m_repeats);
				mem->poke(addr + 2, m_repeats);
				mem->poke(addr + 3, 256 - (offset + 4));
			}
			return offset + 4;
		case TUNECALL:
			if (mem)
			{
				mem->poke(addr,     TUNECALL);
				mem->pokeW(addr + 1,(m_offset + base));
			}
			return 3;
		case TUNERET:
			if (mem) mem->poke(addr, TUNERET);
			return 1;
		case TUNESTOP:
			if (mem) mem->poke(addr, TUNESTOP);
			return 1;
		default:
			if (mem) 
			{
				mem->poke(addr, m_type);
				if (isNote())
					mem->poke(addr + 1, m_duration);
				else	mem->poke(addr + 1, m_effect);
			}
			return 2;
	}
}

Tune128::Tune128()
{
}

Tune128::~Tune128()
{
}

/* Find the highest common factor of i and j by brute force 
 *
 * Although this gets called repeatedly, it shouldn't be too 
 * inefficient. Odds are that after it's been called a couple of times,
 * i will already be the correct hcf and so the first divisibility check
 * will get us out of the while loop.
 *
 */ 
static int hcf(int i, int j)
{
	int p;

	if (i == -1) return j;
	if (j == -1) return i;
	if (i == 0 || j == 0 || i == 1 || j == 1) return 1;

	if (i < j) p = i; else p = j;
	while (p >= 1)
	{
		if (((i/p)*p) == i && ((j/p)*p) == j) return p;
		--p;
	}
	return 1;
}

int NoteSequence::getHcf(int soFar)
{
	for (int m = 0; m < m_count; m++)
	{
		if (!m_data[m].isNote()) continue;
		soFar = hcf(soFar, m_data[m].m_duration);
	}
	return soFar;
}

int Tune128::getTimeBase()
{
	int mhcf= - 1;
	for (int n = 0; n < CHANNELS; n++)
	{
		mhcf = m_data[n].getHcf(mhcf);
	}
	return mhcf;
}


Tune128& Tune128::operator = (Tune128 &other)
{
	for (int n = 0; n < CHANNELS + 1; n++)
	{
		m_data[n].clear();
		m_data[n].append(other.m_data[n]);
	}
}

void Tune128::append(int channel, Note128 n)
{
	if (channel < 0 || channel >= CHANNELS + 1) return;

	m_data[channel].append(n);
}

void Tune128::load(SpectrumMemory *mem, int addr)
{
	int base;

	for (int n = 0; n < CHANNELS; n++)
	{
		base = mem->peekW(addr);
		load(n, mem, base, 0xFF, 0x10000);
		addr += 2;
/* DEBUGGING CODE */
		printf("Channel %d at %x:\n", n, base);
		m_data[n].dump(stdout);
		printf("\n");

	}

}

void Tune128::load(int channel, SpectrumMemory *mem, 
		unsigned base, int term, unsigned top)
{
	jswByte b;
	Note128 n;
	unsigned rel;
	int count;

//	printf("Decode chunk at %x term=%d top=%x\n", base, term, top);
	while (base < top)
	{
		b = mem->peek(base);
		if (b == term) return;
//		printf("%x: Decode %x\n", b, base);
		++base;
		switch(b)
		{
			case TUNERET: 
			case TUNESTOP: return;

			case TUNEDJNZ:
			base++;
			count=mem->peek(base++);
			rel  =mem->peek(base++);
			if (count == 0) count = 256;
//			printf("count=%d rel=%d\n", count, rel);
			while (count)
			{
				load(channel, mem, base - 0x101 + rel, -1, base-4);	
				--count;
			}
			break;

			case TUNECALL:
			rel = mem->peekW(base);
			base += 2;
			load(channel, mem, rel, TUNERET, 0x10000);
			break;

			default:
			n.m_type = b;
			b = mem->peek(base++);
			if (n.isNote()) n.m_duration = b;
			else		n.m_effect   = b;	
			append(channel, n);
			break;

		}	
	} 
}

/* Find out if a tune can be compressed in any way. */
void Tune128::compress(NoteSequence &src, NoteSequence &dest)
{
	/* Firstly, check for repeated sections. */
	int seqLen;
	int n, m, p;	
	int repeats;
	int oldLen, newLen;
	int srcLen = src.length();
	NoteSequence s1, s2, s3;

	dest.clear();
	for (seqLen = 1; seqLen < (srcLen/2); seqLen++)
	{
		for (n = 0; n < srcLen - 2*seqLen; n++) 
		{
			src.subset(&s1, n, seqLen);
			src.subset(&s2, n + seqLen, seqLen);

			if (s1 == s2)
			{
				/* Possibly compressible */
				repeats = 2;
				while (1)
				{
					src.subset(&s3, n+repeats*seqLen,seqLen);
					if (!(s1 == s3)) break;
					++repeats;
				}
				if (repeats > 255) repeats = 255;

				// Would a DJNZ save anything? 
				oldLen = s1.memLength() * repeats;
				newLen = s1.memLength() + 4;
				if (oldLen > newLen)
				{
printf("Compressing using DJNZ at n=%d: [", n);
s1.dump(stdout);
printf("] repeats=%d oldLen=%d newLen=%d\n", repeats, oldLen, newLen);

					/* OK, do a DJNZ */
					src.subset(&dest, 0, n);
					Note128 *tmp = new Note128(s1);
					tmp->m_repeats = repeats;
					dest.append(*tmp);
					delete tmp;
					src.subset(&s3, n+repeats * seqLen);
					dest.append(s3);
					return;
				}
			}
		}
	}
	printf("No further DJNZ compression possible\n");
// XXX Now try subroutines. We need to try existing subroutines and new
// ones; this probably means a separate variable in which the 
// subroutines are built.
//
	for (n = 0; n < m_data[CHANNELS].length(); n++)
	{
		seqLen = 0;
		while (m_data[CHANNELS].getData()[n + seqLen].m_type != TUNERET) 
			++seqLen;
		if (seqLen == 0) continue;

		for (m = 0; m < srcLen; m++)
		{
			m_data[CHANNELS].subset(&s1, n, seqLen);
			src.subset(&s2, m, seqLen);
			if (s1 == s2)
			{
				oldLen = s1.memLength();
				newLen = 4;
				if (newLen < oldLen)
				{
printf("Compressing using existing CALL at m=%d: ", m);
s1.dump(stdout);
printf(" oldLen=%d newLen=%d\n", oldLen, newLen);

					/* OK, do a CALL */
					src.subset(&dest, 0, m);
					Note128 tmp;
					tmp.m_type = TUNECALL;
				     	tmp.m_offset = m_data[CHANNELS].length();
					dest.append(tmp);
					src.subset(&s3, m + seqLen);
					dest.append(s3);
					return;
				}
			}
		}
	}
/* Look for new repeated sequences which could be converted to CALLs */

	for (seqLen = srcLen / 2; seqLen >= 2; seqLen--)
	{
		for (n = 0; n < srcLen - seqLen; n++)
		{
			src.subset(&s1, n, seqLen);
			for (m = n + seqLen; m < srcLen; m++)
			{
				src.subset(&s2, m, seqLen);
				if (s1 == s2)
				{
					oldLen = 2 * s1.memLength();
					newLen = 4 + s1.memLength();
					if (newLen < oldLen)
					{
printf("Compressing using new CALL at n=%d m=%d: ", n, m);
s1.dump(stdout);
printf(" oldLen=%d newLen=%d\n", oldLen, newLen);
						src.subset(&dest, 0, n);
						Note128 tmp, tmp2;
						tmp.m_type = TUNECALL;
						tmp.m_offset = m_data[CHANNELS].memLength();
						dest.append(tmp);
						m_data[CHANNELS].append(s1);
						tmp2.m_type = TUNERET;
						m_data[CHANNELS].append(tmp2);
						src.subset(&s3, n + seqLen,
							m - (n + seqLen));
						dest.append(s3);
						dest.append(tmp);
						src.subset(&s3, m + seqLen);
						dest.append(s3);

						return;
					}
				}
			}
		}
	}
	printf("No further compression possible\n");
}



int Tune128::store(int channel, SpectrumMemory *mem, unsigned base,
		unsigned addr)
{
	int seqlen, m, n, len1, len2;
	NoteSequence buf1, buf2;

	buf2.clear();	
	buf2.append(m_data[channel]);
	do
	{
		buf1.clear();
	       	buf1.append(buf2);
		len1 = buf1.memLength();
		compress(buf1, buf2);
		len2 = buf2.memLength();
	} while (len2 < len1);

	return buf2.store(mem, base, addr);
}

int Tune128::store(SpectrumMemory *mem, int base, int maxlen)
{
	int n;
	int reqlen = 6;
	int len[CHANNELS];
	unsigned addr[CHANNELS];

	m_data[CHANNELS].clear();
      	for (n = 0; n < CHANNELS; n++)
	{
		printf("Channel %d:\n", n);
		m_data[n].dump(stdout);
		printf("\n");
		len[n] = store(n, NULL, base + 6, 0);
		reqlen += len[n];
	}
	reqlen += m_data[CHANNELS].memLength();
	for (n = 0; n < CHANNELS; n++)
	{
		addr[n] = base + reqlen;
		reqlen += len[n];
	}
	fflush(stdout);
	if (reqlen > maxlen) return 0;
	m_data[CHANNELS].store(mem, base + 6, base + 6);
      	for (n = 0; n < CHANNELS; n++)
	{
		mem->pokeW(base + 2*n,  addr[n]);
		store(n, mem, base + 6, addr[n]); 
	}
	return 1;	
}

void Tune128::getChannel(NoteSequence *result, int ch, int starttime)
{
	int n, pos;
	const Note128 *b;

	if (ch < 0 || ch > CHANNELS) 
	{
		result->clear();
		return;
	}

	n = m_data[ch].length();
	b = m_data[ch].getData();
	pos = 0;

	if (b) while (starttime > 0)
	{
		if (b[pos].isNote())
		{
			starttime -= b[pos].m_duration;
		}
		++pos;	
		n--;
		if (n <= 0) 
		{
			result->clear();
			return;
		}
	}
	m_data[ch].subset(result, pos);
}



